//###########################################################
// File: mmc_spi.c
//
// Read-/Writeroutines for MMC MultiMedia cards and
// SD SecureDigital cards in SPI mode.
//
// This will work only for MMC cards with 512 bytes block length !
//
// 30.12.2006 Started ARM support.
//
// 10.10.2006 Added a busy check after CMD0. Without this one of
//            my newer 512MB NoName card hangs.
//
// 16.09.2006 Changed MMCCommand(). Should wait for response !
//            Changed MMCReadSector(), MMCWriteSector(). 
//
// 08.09.2006 Removed one dummy write in MMCWriteSector()
//
// 04.09.2006 Made new MMC_IO_Init(). Removed IO pin settings from
//            MMCIdentify(). Why ? See comments above MMC_IO_Init(). 
//
// 20.06.2006 Store value for SPI speed and switch to double speed
//            for Read/WriteSector(). Restore SPI speed at end of
//            routines.
//
// 22.08.2005 Bugfix for SanDisk 256MB SD card in MMCIdentify();
//
// 29.09.2004 split SPI_WRITE() into SPI_WRITE() and SPI_WAIT()
//            speeds up program because CPU can do something else
//            while SPI hardware module is shifting in/out data 
//            see MMCReadSector() and MMCWriteSector()
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
//#########################################################################
// Last change: 30.12.2006
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: WinAVR 3.4.5
// Compiler: WinARM
//#########################################################################
#if defined(__arm__) || defined(__thumb__)
#error "Sorry no code for arm here !"
#endif

#if defined(__AVR__) || defined(ATMEGA)
 #include <avr/io.h>
#endif

#include "mydefs.h"
#include "dos.h"

#if defined (MMC_DEBUG_IDENTIFY) || defined (MMC_DEBUG_SECTORS) || defined (MMC_DEBUG_COMMANDS)
 #include "serial.h" // For debug only
 #include "printf.h" // For debug only
#endif

#ifdef MMC_CARD_SPI

//######################################################
unsigned char MMCCommand(unsigned char command, unsigned long adress)
//######################################################
{
 unsigned char timeout,by;

#ifdef MMC_DEBUG_COMMANDS
 printf("Cmd %x:",(U16)command);
#endif

 SPI_WRITE(DUMMY_WRITE); 		// Dummy write
 SPI_WAIT();
 SPI_WRITE(command);
 SPI_WAIT();
 SPI_WRITE((unsigned char)((adress & 0xFF000000)>>24)); // MSB of adress
 SPI_WAIT();
 SPI_WRITE((unsigned char)((adress & 0x00FF0000)>>16));
 SPI_WAIT();
 SPI_WRITE((unsigned char)((adress & 0x0000FF00)>>8));
 SPI_WAIT();
 SPI_WRITE((unsigned char)(adress & 0x000000FF));       // LSB of adress
 SPI_WAIT();
 SPI_WRITE(0x95); 			 // Checksum for CMD0 GO_IDLE_STATE and dummy checksum for other commands
 SPI_WAIT();                             

#ifdef MMC_DEBUG_COMMANDS
   printf(" %x",(U16)SPI_DATA_REGISTER);
#endif

 timeout=0;
 by=DUMMY_WRITE;
 
 //wait for response
 while(by==DUMMY_WRITE)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_DATA_REGISTER;
   
#ifdef MMC_DEBUG_COMMANDS
   printf(" %x",(U16)by);
#endif
   timeout++;
   if(timeout>250) break; // no response
  }

 return by;      			 
}

//######################################################
unsigned char MMCReadSector(unsigned long sector, unsigned char *buf)
//######################################################
{
 U16 i;
 unsigned char by;
 unsigned long startadr;
 unsigned char  tmpSPSR;

#ifdef MMC_DEBUG_SECTORS
 printf("RS %lu 0x%08lX\n",sector,sector);
#endif

 if(sector>=maxsect) return 1; //sectornumber too big

#if defined(__arm__) || defined(__thumb__)
 tmpSPSR=SPI_SPEED_REGISTER;
 SPI_SPEED_REGISTER=8;

#elif defined(__AVR__) || defined(ATMEGA)

 //tmpSPCR=SPCR;
 tmpSPSR=SPSR; //save old value
 SPSR=0x01;    //give me double speed

#endif

 MMC_CS_OFF();

//calculate startadress of the sector
 //This will work only up to 4GB
 startadr=sector * BYTE_PER_SEC;

 by=MMCCommand(MMC_READ_BLOCK,startadr);

 while(by!=START_BLOCK_TOKEN)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_DATA_REGISTER; // wait for card response

#ifdef MMC_DEBUG_COMMANDS
// no way to come out of this :( skip this sector !?
     if(by==0x01) // ERROR !
      {
       // One of my SD cards sends this error. My cardreader has also
       // problems to read (NOT write !) the card completely. Trash. 
       printf("\nRead error 0x01 at sector %lu !\n",sector);

       MMC_CS_ON();

       #if defined(__arm__) || defined(__thumb__)
        SPI_SPEED_REGISTER=tmpSPSR;
       #elif defined(__AVR__) || defined(ATMEGA)
        SPSR=tmpSPSR;    //restore old value
       #endif

       // data buffer is not valid at this point !
       return 1;
      }
#endif      
  }

#ifdef STANDARD_SPI_READ
 for(i=0; i< BYTE_PER_SEC; i++)
  {
   SPI_WRITE(DUMMY_WRITE); // start shift in next byte
   SPI_WAIT();      // wait for next byte
   *buf++ =SPI_DATA_REGISTER;       // store byte in buffer
  }
  
 SPI_WRITE(DUMMY_WRITE); // shift in crc part1
 SPI_WAIT();
 SPI_WRITE(DUMMY_WRITE); // shift in crc part2
 SPI_WAIT();
#endif // STANDARD_SPI_READ

#ifdef FAST_SPI_READ
// The following code looks very strange !
// The idea is not to stop the cpu while SPI module transfers data.
// You have 16 cpu cycles until transmission has finished !
// You can use this time to do something like storing your last data
// or get your next data out of memory, doing some loop overhead.....
// Don't wait for end of transmission until you have done something better ;)
//
 SPI_WRITE(DUMMY_WRITE); // shift in first byte
 SPI_WAIT();      // we have to wait for the first byte, but ONLY for the first byte
 by=SPI_DATA_REGISTER;         // get first byte, but store later !
 
 SPI_WRITE(DUMMY_WRITE); // start shift in next byte

 i=BYTE_PER_SEC-1;                 
 while(i) //execute the loop while transmission is running in background
  {
   *buf++ =by;         // store last byte in buffer while SPI module shifts in new data
   i--;
   SPI_WAIT();      // wait for next byte
   by=SPI_DATA_REGISTER;         // get next byte, but store later !
   SPI_WRITE(DUMMY_WRITE); // start shift in next byte
   // do the for() loop overhead at this point while SPI module shifts in new data
  } 

 // last SPI_WRITE(DUMMY_WRITE); is shifting in crc part1 at this point
 *buf =by;         // store last byte in buffer while SPI module shifts in crc part1
 SPI_WAIT();
 SPI_WRITE(DUMMY_WRITE); // shift in crc part2
 SPI_WAIT();
#endif //FAST_SPI_READ

 MMC_CS_ON();

#if defined(__arm__) || defined(__thumb__)
 SPI_SPEED_REGISTER=tmpSPSR;

#elif defined(__AVR__) || defined(ATMEGA)

 SPSR=tmpSPSR;    //restore old value

#endif

 return 0;
}

#ifdef DOS_WRITE
//######################################################
unsigned char MMCWriteSector(unsigned long sector, unsigned char *buf)
//######################################################
{
 U16 i;
 unsigned char by;
 unsigned long startadr;
 unsigned char tmpSPSR;


#ifdef MMC_DEBUG_SECTORS
 printf("WS %lu 0x%08X\n",sector,sector);
#endif

 if(sector>=maxsect) return 1; //sectornumber too big
 
#if defined(__arm__) || defined(__thumb__)
 tmpSPSR=SPI_SPEED_REGISTER;
 SPI_SPEED_REGISTER=8;

#elif defined(__AVR__) || defined(ATMEGA)

 //tmpSPCR=SPCR; //we should save this also !
 tmpSPSR=SPSR; //save old value
 SPSR=0x01;    //give me double speed

#endif

 MMC_CS_OFF();

//calculate startadress
 //This will work only up to 4GB
 startadr=sector * BYTE_PER_SEC;

 MMCCommand(MMC_WRITE_BLOCK,startadr);

 SPI_WRITE(DUMMY_WRITE);  // Dummy write. Do we need this ?
 SPI_WAIT();

#ifdef STANDARD_SPI_WRITE
 SPI_WRITE(START_BLOCK_TOKEN); // start block token for next sector
 SPI_WAIT();
                  
 for(i=0; i<BYTE_PER_SEC; i++)
  {
   SPI_WRITE(*buf++); // shift out next byte
   SPI_WAIT();    // wait for end of transmission
  }
#endif //STANDARD_SPI_WRITE

#ifdef FAST_SPI_WRITE
 SPI_WRITE(START_BLOCK_TOKEN); // start block token for next sector

 i=BYTE_PER_SEC;                 
 while(i) // execute the loop while transmission is running in background
  {
   // do the for() loop overhead at this point while SPI module shifts out new data
   by = *buf++;       // get next data from memory while SPI module shifts out new data
   i--;
   SPI_WAIT();    // wait for end of transmission
   SPI_WRITE(by); // start shift out next byte
  }

 SPI_WAIT();      // wait til last byte is written to MMC
#endif //FAST_SPI_WRITE

 SPI_WRITE(DUMMY_WRITE); // 16 bit crc follows data
 SPI_WAIT();
 SPI_WRITE(DUMMY_WRITE);
 SPI_WAIT();

 SPI_WRITE(DUMMY_WRITE); // read response
 SPI_WAIT();
 by=SPI_DATA_REGISTER & 0x1F;
 if(by != 0x05)   // data block accepted ?
  {
#ifdef MMC_DEBUG_COMMANDS
       printf("\nWrite error at sector %lu !\n",sector);
#endif      

   MMC_CS_ON();

   #if defined(__arm__) || defined(__thumb__)
    SPI_SPEED_REGISTER=tmpSPSR;
   #elif defined(__AVR__) || defined(ATMEGA)
    SPSR=tmpSPSR;    //restore old value
   #endif

   return 1;
  }

 do
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
// busy is not a byte value ! card sends 0 BITS if busy. So it can be any of this:
// 00000000 busy
// 00000001 busy has gone
// 00000011 busy has gone
// 00000111 busy has gone
// 00001111 busy has gone
// 00011111 busy has gone
// 00111111 busy has gone
// 01111111 busy has gone
  }while(SPI_DATA_REGISTER == 0x00); // wait til busy is gone

 MMC_CS_ON();

#if defined(__arm__) || defined(__thumb__)
 SPI_SPEED_REGISTER=tmpSPSR;
 
#elif defined(__AVR__) || defined(ATMEGA)

 SPSR=tmpSPSR;    //restore old value

#endif

 return 0;
}
#endif //DOS_WRITE

//######################################################
// Init MMC IO pins
//
// Removed this from MMCIdentify().
// Maybe some other devices are connected to SPI bus.
// All chip select pins of these devices should have
// high level before starting SPI transmissions !
// So first call MMC_IO_Init(), after that for example
// VS1001_IO_Init(), SPILCD_IO_Init() and AFTER that MMCIdentify().
void MMC_IO_Init(void)
//######################################################
{
#if defined(__arm__) || defined(__thumb__)

//    IODIR0 |= SPI0_SEL; // maybe we want to use SPI1
//    IOSET0 = SPI0_SEL;  // and not SEL as chip select
 MMC_DDR_CS|=MMC_BIT_CS;
 MMC_CS_ON();

 SPI_CTRL_REGISTER = 0x00; //Disable SPI
 SPI_PINSEL &= SPI_PINSEL_CLEAN_MASK;
 SPI_PINSEL |= SPI_PINSEL_SET_MASK;

#elif defined(__AVR__) || defined(ATMEGA)
 sbi(MMC_CS_PORT,MMC_CS_BIT); //MMC_CS set 1
 sbi(MMC_CS_DDR,MMC_CS_BIT);  //MMC_CS output

 cbi(MMC_SCK_PORT,MMC_SCK_BIT); //SCK set 0
 sbi(MMC_SCK_DDR,MMC_SCK_BIT);  //SCK output

 cbi(MMC_MISO_DDR,MMC_MISO_BIT);  //MISO input

 cbi(MMC_MOSI_PORT,MMC_MOSI_BIT); //MOSI set 0
 sbi(MMC_MOSI_DDR,MMC_MOSI_BIT);  //MOSI output
#else
#  error "processor type not defined in mmc.c -> MMC_IO_Init()"
#endif
}

//######################################################
unsigned char MMCIdentify(void)
//######################################################
{
 unsigned char by;
 U16 i;
 U16 c_size, c_size_mult;
// U16 read_bl_len;
// unsigned long drive_size;

// Init SPI with a very slow transfer rate first !

#if defined(__arm__) || defined(__thumb__)

    SPI_SPEED_REGISTER = 128;
    SPI_CTRL_REGISTER = SPI_MSTR;

#elif defined(__AVR__) || defined(ATMEGA)
// SPCR SPI Controlregister
// SPIE=0; //No SPI Interrupt
// SPE=1;  //SPI Enable
// DORD=0; //Send MSB first
// MSTR=1; //I am the master !
// CPOL=0; //SCK low if IDLE
// CPHA=0; //SPI Mode 0
// SPR1=1; //SPI Clock = f/128 = 125kHz @16MHz Clock
// SPR0=1; //or f/64 if SPI2X = 1 in SPSR register
 SPCR=0x53;

// SPSR SPI Statusregister
// SPI2X=1; //Double speed for SPI = 250kHz @16MHz Clock
// SPSR=0x01;
 SPSR=0x00;
#else
#  error "processor type not defined in mmc.c -> MMCIdentify()"
#endif

 for(i=0; i<10; i++)
  {
   SPI_WRITE(DUMMY_WRITE); // give min 74 SPI clock pulses before
                           // sending commands
   SPI_WAIT();
  }                                      
 
 MMC_CS_OFF();

#ifdef MMC_DEBUG_IDENTIFY
 puts("Send CMD0\n");
#endif

//send CMD0 for RESET
 by=MMCCommand(MMC_GO_IDLE_STATE,0);

 if(by==0xFF)  // no response from card
  {
   MMC_CS_ON();
   return 1;
  }

// busy is not a byte value ! card sends 0 BITS if busy. So it can be any of this:
// 00000000 busy
// 00000001 busy has gone
// 00000011 busy has gone
// 00000111 busy has gone
// 00001111 busy has gone
// 00011111 busy has gone
// 00111111 busy has gone
// 01111111 busy has gone
// One of my NoName cards sends 0x00, 0x7F at power on, but 0xFF, 0xFF, 0x01 if reset only with power already on.
// Added a busy check below and now it works !
 while(by==0x00) //wait til busy is gone
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_DATA_REGISTER;
#ifdef MMC_DEBUG_IDENTIFY
   printf(" %x",(U16)SPI_DATA_REGISTER);
#endif
  } 

#ifdef MMC_DEBUG_IDENTIFY
 puts("Send CMD1\n");
#endif

// Repeat CMD1 til result=0
 do
  {
   by=MMCCommand(MMC_INIT,0);
  }while(by!=0);

#ifdef MMC_DEBUG_IDENTIFY
 printf("CMD1 done\n");
#endif

// Read CID
// MMCCommand(MMC_READ_CID,0); // nothing really interesting here

#ifdef MMC_DEBUG_IDENTIFY
 puts("Read CSD\n");
#endif

// Read CSD Card Specific Data
 by=MMCCommand(MMC_READ_CSD,0);

// This waiting loop is for SanDisk 256MB SD
// Works with my other cards too
 while(by!=START_BLOCK_TOKEN)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_DATA_REGISTER; // Wait for card response
  }
   
 for(i=0; i<16; i++) // CSD has 128 bits -> 16 bytes
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_DATA_REGISTER;
   iob[i]=by;
  }

 SPI_WRITE(DUMMY_WRITE); // 16 bit crc follows data
 SPI_WAIT();
 SPI_WRITE(DUMMY_WRITE);
 SPI_WAIT();

#ifdef MMC_DEBUG_IDENTIFY
 printf("Read CSD done\n");
#endif
// Here comes the hard stuff !
// Calculate disk size and number of last sector
// that can be used on your mmc/sd card

 c_size=iob[6] & 0x03; //bits 1..0
 c_size<<=10;
 c_size+=(U16)iob[7]<<2;
 c_size+=iob[8]>>6;

// read_bl_len is always 512 !!??
/*
 by= iob[5] & 0x0F;
 read_bl_len=1;
 read_bl_len<<=by;
*/

 by=iob[9] & 0x03;
 by<<=1;
 by+=iob[10] >> 7;
 
 c_size_mult=1;
 c_size_mult<<=(2+by);

// drive_size=(unsigned long)(c_size+1) * (unsigned long)c_size_mult * (unsigned long)read_bl_len;
// maxsect= drive_size / BYTE_PER_SEC; // Got it !

 maxsect=(unsigned long)(c_size+1) * (unsigned long)c_size_mult;

 MMC_CS_ON();

#ifdef MMC_DEBUG_IDENTIFY
 printf("c_size %u\n",c_size);
 printf("c_size_mult %u\n",c_size_mult);
// printf("read_bl_len %u\n",read_bl_len); // always 512
 printf("DriveSize %lu kB\n",(maxsect>>1));
 printf("maxsect %lu\n",maxsect);
#endif

#if defined(__arm__) || defined(__thumb__)
// Switch to high speed SPI
 SPI_SPEED_REGISTER = 0x8;

#elif defined(__AVR__) || defined(ATMEGA)
// Switch to high speed SPI
// SPR1=0; //SPI Clock = f/4 = 4MHz @16MHz Clock
// SPR0=0; //or f/2 if SPI2X = 1 in SPSR register
 SPCR=0x50;

// SPSR SPI Statusregister
// SPI2X=1; //Double speed for SPI = 8MHz @16MHz Clock
 SPSR=0x01;

#else
#  error "processor type not defined in mmc.c -> MMCIdentify()"
#endif

 return 0;
}


#endif //MMC_CARD_SPI
